unit VerRes;

{************************************************************************
*                                                                       *
*                    Application Versioning Routines                    *
*                   Copyright (C) 1997, John Zaitseff                   *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@unsw.edu.au>
  Date:      20th December, 1997.
  Version:   1.3

  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.

  This file provides an object to access the VERSIONINFO resource in any
  executable file.  In addition, a default object "AppVerRes" is created,
  which returns version information on the currently running executable.

  TVerRes.Create(Filename)  creates the object, reading in the VERSIONINFO
      resource for the specified filename.  If Filename does not point
      to a valid VERSIONINFOable file (EXE, DLL, SYS, etc), an exception
      is raised.
  TVerRes.Free  destroys the object.
  TVerRes.Valid  returns True if the executable contains a valid
      VERSIONINFO resource.
  TVerRes.FixedInfo  returns the language-independent portion of the
      VERSIONINFO resource.
  TVerRes.ResourceString(Res)  returns the language-specific resource
      string in the VERSIONINFO resource.  For example,
      ResourceString('FileVersion') may return something like '1.2'.
  TVerRes.LanguageID  returns the language identifier used when accessing
      language-specific resources.  See below for how this is calculated.
  TVerRes.CompanyName
  TVerRes.FileDescription
  TVerRes.FileVersion
  TVerRes.ProductName
  TVerRes.ProductVersion
  TVerRes.InternalName
  TVerRes.OriginalFilename
  TVerRes.LegalCopyright
      These functions return the language-dependent strings associated with
      the like-named VERSIONINFO resources.  These are simply "short-cuts" to
      using the ResourceString function (eg, calling TVerRes.CompanyName is
      the same as calling TVerRes.ResourceString('CompanyName') ).

  Note that the Win32 functions used to implement these routines state that
  they only work on Win32 files, NOT Win16 files.  Although not documented,
  however, using TVerRes with 16-bit EXEs and DLLs DOES seem to work.

  The algorithm used to determine which language to use for the language-
  dependent resources is as follows:
      1.  Read in the VarFileInfo\Translation part of the VERSIONINFO
          resource.  This is a table of language-character set pairs of
          languages present in the language-dependent part of the
          VERSIONINFO resource.
      2.  If the user-specified language is present (via GetUserDefaultLangID),
          use that language.
      3.  If the user-specified language (minus the sub-language ID) matches
          one present (minus the sub-language ID), use that language.
      4.  If the system default language is present (via
          GetSystemDefaultLangID), use that language.
      5.  If the system default language (minus the sub-language ID) matches
          one present (minus the sub-language ID), use that language.
      6.  Failing all of the above, use the first language specified in the
          translation table.
}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only (v2 or later).'
{$ENDIF}

{$IFNDEF OPTDEBUG}
{$DEFINE DEBUG}
{$ENDIF}

{$IFDEF DEBUG}
    {$B-,D+,I+,J-,L+,O+,P-,Q+,R+,T+,U+,V+,X+,Y+,Z1}
    {$WARNINGS ON}
    {$HINTS ON}
{$ELSE}
    {$B-,D-,I+,J-,L-,O+,P-,Q-,R-,T+,U+,V+,X+,Y-,Z1}
    {$WARNINGS ON}
    {$HINTS ON}
{$ENDIF}


interface

uses
  SysUtils;


type
  { Language-independent version information.  See the Win32 description
    of VS_FIXEDFILEINFO.  Note that FileDateMS and FileDateLS seem to be
    always set to 0 }
  TFixedInfo = record
                 FileVersionMS,
                 FileVersionLS,
                 ProductVersionMS,
                 ProductVersionLS,
                 FileFlagsMask,
                 FileFlags,
                 FileOS,
                 FileType,
                 FileSubtype,
                 FileDateMS,
                 FileDateLS       : integer;
               end;

  { Executable VERSIONINFO resource object definition }
  TVerRes = class
              private
                FVerResource       : pointer;

                FLangID            : word;
                FCharSetID         : word;

                FBlockPrefix       : string;

                FFixedInfo         : TFixedInfo;
                FFixedInfoPresent  : boolean;

                FCompanyName,
                FFileDescription,
                FFileVersion,
                FProductName,
                FProductVersion,
                FInternalName,
                FOriginalFilename,
                FLegalCopyright    : string;

                procedure GetLangTransVals;
                function GetResStr (const Res : string; var ResBuf : string) : string;

              public
                constructor Create (Filename : string);
                destructor Destroy; override;

                function Valid : boolean;
                function LanguageID : word;

                function FixedInfo : TFixedInfo;
                function ResourceString (const Res : string) : string;

                function CompanyName      : string;
                function FileDescription  : string;
                function FileVersion      : string;
                function ProductName      : string;
                function ProductVersion   : string;
                function InternalName     : string;
                function OriginalFilename : string;
                function LegalCopyright   : string;
            end;

  { Exception object for TVerRes object }
  EVerResExc = class(Exception)
                 private
                   FErrorCode : integer;

                 public
                   constructor Create (const Msg : string);
                   constructor CreateErrorCode (const Msg : string;
                                                ErrorCode : integer);
                   property ErrorCode : integer read FErrorCode;
               end;

  { EVerResExc has an ErrorCode field, which returns the error code from
    the failed operation.  In particular, the error code is the same
    as GetLastError, unless TVerRes.Valid returns False, in which case
    ErrorCode returns 0 }


var
  AppVerRes : TVerRes;     { Application VERSIONINFO resource }


implementation

uses
  Windows;


const
  MaxResStrLen = 512;   { Maximum length of any VERSIONINFO resource string }


type
  { Language translation table entry }
  TTransVal    = packed record
                   LangID,
                   CharSetID : word
                 end;
  { Array of table entries, large enough for ANY table! }
  TTransValTbl = array [1..MaxInt div 4] of TTransVal;
  PTransValTbl = ^TTransValTbl;


{************************************************************************}
{ Create the TVerRes object by reading in the VERSIONINFO resource from
  the file pointed to by Filename.  If a blank string is passed as the
  filename, the currently executing program or DLL is used instead.
  If the file is not found or is in the incorrect format (ie, not a Win32
  executable), an exception is raised with ErrorCode set to GetLastError.
  If the executable is valid, but no VERSIONINFO resource exists, the
  object is still created, but TVerRes.Valid will return False. }
constructor TVerRes.Create (Filename : string);

var
  N, N2 : dword;

begin
  inherited Create;

  { If filename is blank, use the currently executing module's filename }
  if Filename = '' then
    begin
      SetLength(Filename, MaxResStrLen);
      N := GetModuleFileName(HInstance, PChar(Filename), MaxResStrLen);
      if N > 0 then
        SetLength(Filename, N)
      else
        raise EVerResExc.CreateErrorCode('Could not obtain module filename',
                                         GetLastError)
    end;

  { Get the size of the VERSIONINFO resource }
  N := GetFileVersionInfoSize(PChar(Filename), N2);
  if N > 0 then
    begin
      { Allocate memory and read in the VERSIONINFO resource }
      GetMem(FVerResource, N);
      if not GetFileVersionInfo(PChar(Filename), 0, N, FVerResource) then
        begin
          FreeMem(FVerResource);
          FVerResource := nil;
          raise EVerResExc.CreateErrorCode('Could not read in VERSIONINFO ' +
                                           'resource from "' + Filename + '"',
                                           GetLastError)
        end
    end
  else if (N = 0) and (GetLastError = 0) then
    { No VERSIONINFO resource is present in the file }
    FVerResource := nil
  else
    { An error occurred: abort creation of the object }
    raise EVerResExc.CreateErrorCode('Could not read in VERSIONINFO resource ' +
                                     'from "' + Filename + '"', GetLastError)
end;

{************************************************************************}
{ Destroy the TVerRes object by freeing any associated memory }
destructor TVerRes.Destroy;

begin
  if Assigned(FVerResource) then
    FreeMem(FVerResource);

  inherited Destroy
end;

{************************************************************************}
{ Although the TVerRes object may have been created without a problem,
  the filename that was passed as the Create parameter may not have a
  VERSIONINFO resource associated with it.  This function returns False
  if that is the case, or True if calling other TVerRes functions will
  succeed }
function TVerRes.Valid : boolean;

begin
  Result := Assigned(FVerResource)
end;

{************************************************************************}
{ Internal procedure to TVerRes.  This sets the FLangID, FCharSetID and
  FBlockPrefix internal variables.  See the beginning of this file for
  the algorithm used to determine which language to use for the language-
  dependent resources }
procedure TVerRes.GetLangTransVals;

  { Internal function }
  function ScanTbl (Tbl : PTransValTbl; TblSize, Lang, Mask : word) : boolean;

  var
    I : word;
    
  begin
    Result := False;
    for I := 1 to TblSize do
      if (Tbl^[I].LangID and Mask) = (Lang and Mask) then
        with Tbl^[I] do
          begin
            FLangID := LangID;
            FCharSetID := CharSetID;
            Result := True;
            break
          end
  end;

var
  Tbl   : PTransValTbl;
  N     : uint;

begin
  if not Valid then
    raise EVerResExc.Create('No VERSIONINFO resource present');

  { If FBlockPrefix is already non-blank, this function has been called
    and there is no need to call it again }
  if FBlockPrefix <> '' then
    exit;

  if VerQueryValue(FVerResource, '\VarFileInfo\Translation', Pointer(Tbl), N) and
     Assigned(Tbl) and (N > 0) then
    begin
      { T points to language translation table.  N is the size of this table
        in bytes (records, after the next line) }
      N := N div sizeof(TTransVal);

      { Search for correct language to use:  user default, user-default
        (minus sub-language), system default, system default (minus sub-
        language), first language in table }
      if not ScanTbl(Tbl, N, GetUserDefaultLangID, $FFFF) then
        if not ScanTbl(Tbl, N, GetUserDefaultLangID, $03FF) then
          if not ScanTbl(Tbl, N, GetSystemDefaultLangID, $FFFF) then
            if not ScanTbl(Tbl, N, GetSystemDefaultLangID, $03FF) then
              begin
                FLangID := Tbl^[1].LangID;
                FCharSetID := Tbl^[1].CharSetID
              end
    end
  else
    begin
      { Language table not present, assume US English }
      FLangID := $0409;
      FCharSetID := 1252
    end;

  FBlockPrefix := Format('\StringFileInfo\%4.4X%4.4X\', [FLangID, FCharSetID])
end;

{************************************************************************}
{ Return the language identifier used for retrieving the language-dependent
  portions of the VERSIONINFO resource }
function TVerRes.LanguageID : word;

begin
  GetLangTransVals;
  Result := FLangID
end;

{************************************************************************}
{ Return the language-independent portion (ie, fixed file information) of
  the VERSIONINFO resource.  This information is cached }
function TVerRes.FixedInfo : TFixedInfo;

var
  FFI : Pointer;
  N   : uint;

begin
  if not Valid then
    raise EVerResExc.Create('No VERSIONINFO resource present');

  { If information has not been cached, retrieve and store it }
  if not FFixedInfoPresent then
    begin

      { Get the fixed file information from the VERSIONINFO resource }
      if not VerQueryValue(FVerResource, '\', FFI, N) then
        raise EVerResExc.CreateErrorCode('Could not read in fixed file ' +
                                         'information from VERSIONINFO resource',
                                         GetLastError)
      else if not Assigned(FFI) then
        raise EVerResExc.Create('No fixed file information present in ' +
                                'the VERSIONINFO resource')
      else
        begin
          { Fixed file information found; cache it in FFixedInfo }
          with FFixedInfo do
            begin
              FileVersionMS := TVSFixedFileInfo(FFI^).dwFileVersionMS;
              FileVersionLS := TVSFixedFileInfo(FFI^).dwFileVersionLS;
              ProductVersionMS := TVSFixedFileInfo(FFI^).dwProductVersionMS;
              ProductVersionLS := TVSFixedFileInfo(FFI^).dwProductVersionLS;
              FileFlagsMask := TVSFixedFileInfo(FFI^).dwFileFlagsMask;
              FileFlags := TVSFixedFileInfo(FFI^).dwFileFlags;
              FileOS := TVSFixedFileInfo(FFI^).dwFileOS;
              FileType := TVSFixedFileInfo(FFI^).dwFileType;
              FileSubtype := TVSFixedFileInfo(FFI^).dwFileSubtype;
              FileDateMS := TVSFixedFileInfo(FFI^).dwFileDateMS;
              FileDateLS := TVSFixedFileInfo(FFI^).dwFileDateLS
            end;
          FFixedInfoPresent := True
        end
    end;

  Result := FFixedInfo
end;

{************************************************************************}
{ Return the value stored in ResBuf if ResBuf is not '', or else return
  the value stored in the language-dependent portion of the VERSIONINFO
  resource that has the keyword matching Res.  If no such keyword exists,
  a blank string is returned.  ResBuf is updated with the returned value }
function TVerRes.GetResStr (const Res : string; var ResBuf : string) : string;

var
  N        : uint;
  P        : PChar;
  SubBlock : string;

begin
  if not Valid then
    raise EVerResExc.Create('No VERSIONINFO resource present');

  if ResBuf <> '' then
    begin
      Result := ResBuf;
      exit
    end;

  GetLangTransVals;
  SubBlock := FBlockPrefix + Res;

  N := MaxResStrLen;
  SetLength(Result, N);

  if VerQueryValue(FVerResource, PChar(SubBlock), Pointer(P), N) then
    begin
      Result := P;
      ResBuf := Result
    end
  else
    Result := ''
end;


{************************************************************************}
{ Return the value stored in the language-dependent portion of the
  VERSIONINFO resource that has the keyword matching Res.  If no such
  keyword exists, a blank string is returned }
function TVerRes.ResourceString (const Res : string) : string;

var
  S : string;

begin
  S := '';
  Result := GetResStr(Res, S)
end;

{************************************************************************}
{ Return the language-dependent string associated with the standard
  CompanyName VERSIONINFO resource }
function TVerRes.CompanyName : string;

begin
  Result := GetResStr('CompanyName', FCompanyName)
end;

{************************************************************************}
{ Return the language-dependent string associated with the standard
  FileDescription VERSIONINFO resource }
function TVerRes.FileDescription : string;

begin
  Result := GetResStr('FileDescription', FFileDescription)
end;

{************************************************************************}
{ Return the language-dependent string associated with the standard
  FileVersion VERSIONINFO resource }
function TVerRes.FileVersion : string;

begin
  Result := GetResStr('FileVersion', FFileVersion)
end;

{************************************************************************}
{ Return the language-dependent string associated with the standard
  ProductName VERSIONINFO resource }
function TVerRes.ProductName : string;

begin
  Result := GetResStr('ProductName', FProductName)
end;

{************************************************************************}
{ Return the language-dependent string associated with the standard
  ProductVersion VERSIONINFO resource }
function TVerRes.ProductVersion : string;

begin
  Result := GetResStr('ProductVersion', FProductVersion)
end;

{************************************************************************}
{ Return the language-dependent string associated with the standard
  InternalName VERSIONINFO resource }
function TVerRes.InternalName : string;

begin
  Result := GetResStr('InternalName', FInternalName)
end;

{************************************************************************}
{ Return the language-dependent string associated with the standard
  OriginalFilename VERSIONINFO resource }
function TVerRes.OriginalFilename : string;

begin
  Result := GetResStr('OriginalFilename', FOriginalFilename)
end;

{************************************************************************}
{ Return the language-dependent string associated with the standard
  LegalCopyright VERSIONINFO resource }
function TVerRes.LegalCopyright : string;

begin
  Result := GetResStr('LegalCopyright', FLegalCopyright)
end;


{************************************************************************}
{ Construct an exception object for the TVerRes object.  Set the error
  code to 0 }
constructor EVerResExc.Create (const Msg : string);

begin
  inherited Create(Msg);
  FErrorCode := 0
end;

{************************************************************************}
{ Construct an exception object for the TVerRes object, including setting
  an error code }
constructor EVerResExc.CreateErrorCode (const Msg : string; ErrorCode : integer);

begin
  inherited CreateFmt('%s (Error code %d)', [Msg, ErrorCode]);
  FErrorCode := ErrorCode
end;

{************************************************************************}

initialization
  AppVerRes := TVerRes.Create(ParamStr(0))
finalization
  AppVerRes.Free
end.

